import os
import pickle

from .utils import *
from logging import warn


class Sky:

    def __init__(
        self, 
        api_key: Union[str, None] = None, 
        file_path: str = "sky_credentials.json", 
        token_path: Union[str, None]  = None
        ):
        """Blackbaud Sky API client 

        This class uses a :class:`authlib.integrations.requests_client.OAuth2Session` for 
        calls to the Blackbaud Sky API.
        
        """
        self.token = None
        self.client = None
        self.file_path = file_path

        # Seeing if the user saved the api key as an environment variable
        if os.getenv('BB_API_KEY'):
            self.api_key = os.getenv('BB_API_KEY')
        elif api_key:
            self.api_key = api_key
        else:
            warn("""
            A api key is needed to call the Blackbaud sky api. You can either initialize it when calling 
            the Sky class or you can save it in a environment variable called BB_API_KEY
            """)

        # Path to cached token
        if token_path:
            self.token_path = token_path
        elif os.getenv('BB_TOKEN_PATH'):
            self.token_path = os.getenv('BB_TOKEN_PATH')
        else:
            self.token_path = '.sky-token'

    @authorize
    def get(
        self,
        params: Union[dict, None] = None,
        reference: str = 'school',
        endpoint: str = "roles"
        ) -> dict:
        """ Get request to the Sky API
        Args:
            params: Dictionary that defines parameters to be passed to 
            the api
            reference: Which SKY Api refrence are you calling. See them here 
            https://developer.blackbaud.com/skyapi/apis
            endpoint: The specific endpioint that exist in the given api reference

        Returns:
           Dictionary with data from the sky api
        """
        url = self._get_url(reference, endpoint)

        apiCall = GetRequest(
            self.client,
            url,
            self.request_header,
            params=params
        )

        data = apiCall.getData()
        self._saveToken(apiCall.updateToken(self.token))
        return data


    @authorize
    def post(
        self,
        data: dict,
        reference: str = 'school',
        endpoint: str = "users",
        ) -> dict:
        """Post request to the Sky API
        Args:
            data: Dictionary that defines the request data to be passed to 
            the api in order to create a new record
            reference: Which SKY Api refrence are you calling. See them here 
            https://developer.blackbaud.com/skyapi/apis
            endpoint: The specific endpioint that exist in the given api reference

        Returns:
           Dictionary with data from the sky api
        """
        if not data:
            raise ValueError("Request data is required for post requests")
        url = self._get_url(reference, endpoint)

        apiCall = PostRequest(
            self.client,
            url,
            self.request_header,
            data=data
        )

        data = apiCall.getData()
        self._saveToken(apiCall.updateToken(self.token))
        return data


    @authorize
    def patch(
        self,
        reference: str = 'school',
        endpoint: str = "roles",
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        **kwargs
        ) -> dict:
        """Patch requests to the Sky API
        Args:
            data: Dictionary that defines the request data to be passed to 
            the api in order to create a new record
            reference: Which SKY Api refrence are you calling. See them here 
            https://developer.blackbaud.com/skyapi/apis
            endpoint: The specific endpioint that exist in the given api reference

        Returns:
           Dictionary with data from the sky api
        """
        url = self._get_url(reference, endpoint)

        apiCall = PatchRequest(
            self.client,
            url,
            self.request_header,
            params=params,
            data=data
        )

        data = apiCall.getData(**kwargs)
        self._saveToken(apiCall.updateToken(self.token))
        return data


    @authorize
    def delete(
        self,
        reference: str = 'school',
        endpoint: str = "roles",
        params: Union[dict, None] = None,
        data: Union[dict, None] = None,
        **kwargs
        ) -> dict:
        """Delete requests to the sky API

        Args:
            reference: Which SKY Api refrence are you calling. See them here 
            https://developer.blackbaud.com/skyapi/apis
            endpoint: The specific endpioint that exist in the given api reference
            **kwargs: ... Honestly don't know yet. Never used this endpoint. Just 
            adding for testing

        Returns:
           Dictionary with data from the sky api
        """
        url = self._get_url(reference, endpoint)

        apiCall = DeleteRequest(
            self.client,
            url,
            self.request_header,
            params = params,
            data=data
        )

        data = apiCall.getData(**kwargs)
        self._saveToken(apiCall.updateToken(self.token))
        return data


    def _loadCachedToken(self) -> Union[None, OAuth2Token]:
        """Load Sky API token from cache"""
        # Loading token from binary file
        if os.path.exists(self.token_path):
            with open(self.token_path, 'rb') as token:
                self.token = pickle.load(token)          
        return self.token


    def _get_url(self, reference: str, endpoint: str) -> str:
        """Format api requests url

        Args:
            reference: 
            endpoint: 

        Returns:
            API url to call
        """
        return f'https://api.sky.blackbaud.com/{reference}/v1/{endpoint}'


    def _saveToken(
        self, 
        token: OAuth2Token
        ) -> None:
        """Save OAuth2Token for future use"""
        with open(self.token_path,  'wb') as f:
            pickle.dump(token, f)
        self.token = token


    @property
    def request_header(self):
        """API key to pass to Request header"""
        return {"Bb-Api-Subscription-Key": self.api_key}


import webbrowser
import wsgiref.simple_server
import wsgiref.util
import logging
import json
import urllib.parse as urlparse

from typing import Union, Iterable, Type, TypeVar, Generic, Tuple, Mapping, Callable, Any
from authlib.integrations.requests_client import OAuth2Session
from authlib.oauth2.rfc6749.wrappers import OAuth2Token

_LOGGER = logging.getLogger(__name__)

Flow = TypeVar('Flow')

class AuthApp(Generic[Flow]):
    AUTHORIZATION_URL='https://oauth2.sky.blackbaud.com/authorization'


    def __init__(
        self,
        client: OAuth2Session,
        host: Union[str, int],
        port: int,
        client_id: str,
        client_secret: str
        ):
        """OAuth 2.0 Authorization App

        Here's an example of using :class:`AuthApp`::

        from funky.auth import AuthApp

        # Create the flow using the sky credentials file from the with 
        # client_id, client_secret, and redirect URI from your BlackBaud Sky api app
        app = AuthApp.load_credentials("path/to/sky_credentials.json")
        sky_token = app.run_local_server()
        """
        self.client = client
        self.host = host
        self.port = port
        self.client_id = client_id
        self.client_secret = client_secret
        
    @classmethod
    def load_credentials(
        cls, 
        sky_credentials: str
        ) -> Type[Flow]:
        """Creates a :class:`Flow` instance from a sky_credentials file.

        Args:
            sky_credentials: The path to the sky credentials .json file.
            kwargs: Any additional parameters passed to
                :class:`authlib.integrations.requests_client.OAuth2Session`

        Returns:
            Flow: The constructed Flow instance.
        """
        with open(sky_credentials, "r") as json_file:
            credentials = json.load(json_file)

        # Parsing the url for the host and port
        url = urlparse.urlparse("http://localhost:8080")
        host = url.hostname
        port = url.port
        clientID= credentials['client_id']
        clientSecret = credentials['client_secret']

        client = OAuth2Session(
            client_id = clientID,
            client_secret = clientSecret,
            redirect_uri = credentials['redirect_uri'],
            token_endpoint='https://oauth2.sky.blackbaud.com/token',
            authorization_endpoint='https://oauth2.sky.blackbaud.com/authorization',
            token_endpoint_auth_method='client_secret_basic'
        )
        
        return cls(
            client,
            host,
            port,
            clientID,
            clientSecret
        )
    

    def run_local_server(
        self
        ) -> OAuth2Token:
        """Perform authorization flow using local server

        It will start a local web server to listen for the authorization
        response. Once authorization is complete the authorization server will
        redirect the user's browser to the local web server. The web server
        will get the authorization code from the response and shutdown. The
        code is then exchanged for a token.

        Returns:
            class:`authlib.oauth2.rfc6749.wrappers.OAuth2Token`

        """
        wsgi_app = _RedirectWSGIApp("You may close the browser now")
        local_server = wsgiref.simple_server.make_server(
            self.host, self.port, wsgi_app, handler_class=_WSGIRequestHandler
        )

        url, _ = self.authorization_url()

        webbrowser.open(url, new=1, autoraise=True)

        local_server.handle_request()

        # OAuth 2.0 should only occur over https.
        authorization_response = wsgi_app.last_request_uri.replace("http", "https")
        token = self.fetch_token(authorization_response=authorization_response)
        print('all good')
        token['client_id'] = self.client_id
        token['client_secret']  = self.client_secret
        return token
    

    def authorization_url(self) -> Tuple[str, str]:
        """Generates an authorization URL

        This is the first step in the OAuth 2.0 Authorization Flow. The user's
        browser should be redirected to the returned URL.

        This method calls
        :meth:`authlib.integrations.requests_client.OAuth2Session.create_authorization_url`
        and specifies the client configuration's authorization URI. This is required in order to 
        obtain a refresh token.
        """
        return self.client.create_authorization_url(self.AUTHORIZATION_URL)


    def fetch_token(self, **kwargs) -> OAuth2Token:
        """Completes the Authorization Flow and obtains an access token.

        This is the final step in the OAuth 2.0 Authorization Flow. This is
        called after the user consents.

        Args:
        kwargs: Arguments passed through to
            :meth:`authlib.integrations.requests_client.OAuth2Session.fetch_token`. 
            client secret must be specified
        This method calls
        :meth:`authlib.integrations.requests_client.OAuth2Session.fetch_token`
        and specifies the client configuration's token URI
        """
        kwargs.setdefault("client_secret", self.client.client_secret)
        return self.client.fetch_token(self.client.metadata['token_endpoint'], **kwargs)
    
    
class _WSGIRequestHandler(wsgiref.simple_server.WSGIRequestHandler):
    """Custom WSGIRequestHandler.

    Uses a named logger instead of printing to stderr.
    """

    def log_message(self, format, *args):
        # pylint: disable=redefined-builtin
        # (format is the argument name defined in the superclass.)
        _LOGGER.info(format, *args)


class _RedirectWSGIApp(object):
    """WSGI app to handle the authorization redirect.

    Stores the request URI and displays the given success message.
    """
    def __init__(self, success_message):
        """
        Args:
            success_message (str): The message to display in the web browser
                the authorization flow is complete.
        """
        self.last_request_uri = None
        self._success_message = success_message

    def __call__(
        self, 
        environ: Mapping[str, Any], 
        start_response: Callable[[], str]
        ) -> Iterable[bytes]:
        """WSGI Callable.

        Args:
            environ (Mapping[str, Any]): The WSGI environment.
            start_response (Callable[str, list]): The WSGI start_response
                callable.

        Returns:
            Iterable[bytes]: The response body.
        """
        start_response("200 OK", [("Content-type", "text/plain")])
        self.last_request_uri = wsgiref.util.request_uri(environ)
        return [self._success_message.encode("utf-8")]


from .auth import AuthApp, OAuth2Token
from authlib.integrations.requests_client import OAuth2Session
from typing import Union


def authorize(func):
    """ Load OAuth2Session

    A :class:`authlib.integrations.requests_client.OAuth2Session ` is needed in order to 
    call the sky api and save api tokens.
    
    loadClient first attempts to load a token from the users cache but if there's no token 
    available it will automatically launch a local web server in order to authenticate you
    with the Sky API
    """    

    def wrap(*args, **kwargs):
        # Assing the sky instance
        skyObject = args[0]
        
        # Chcking if theres a client existant
        if not skyObject.client:
            # Chccking if the user already has a token caced
            if not skyObject._loadCachedToken():
                authorizationApp(skyObject)
            # Initalizing the client class
            skyObject.client = OAuth2Session(
                    token= skyObject.token,
                    client_id = skyObject.token['client_id'],
                    client_secret = skyObject.token['client_secret'],
                    token_endpoint='https://oauth2.sky.blackbaud.com/token',
                    token_endpoint_auth_method='client_secret_basic'
                )
        # Running the call function
        return func(*args, **kwargs)
    return wrap


def authorizationApp(skyObject) -> None:
    """Launch server to retrieve Sky API token"""
    app = AuthApp.load_credentials(skyObject.file_path)
    skyObject.token = app.run_local_server()
    skyObject._saveToken(skyObject.token)


class BaseRequest:

    def __init__(
        self,
        client: OAuth2Session,
        url: str,    
        header: dict,
        params: Union[dict, None] = None,
        data: Union[dict, None] = None
        ):
        self.client = client
        self.url = url
        self.header = header
        self.params = params
        self.data = data

    def getData(self):
        pass

    def cleanData(self):
        pass

    def updateToken(self, cache_token):
        self.client.token['client_id'] = cache_token['client_id']
        self.client.token['client_secret'] = cache_token['client_secret']
        if self.client.token != cache_token:
            return self.client.token
        return cache_token


class GetRequest(BaseRequest):

    def getData(self):
        data = self.client.get(
            self.url, 
            headers = self.header, 
            params=self.params)
        return data.json()

    def cleanData(self):
        pass


class PostRequest(BaseRequest):

    def getData(self):
        data = self.client.post(
            self.url, 
            headers = self.header, 
            data=self.data)
        return data

    def cleanData(self):
        pass


class PatchRequest(BaseRequest):

    def getData(self, **kwargs):
        data = self.client.patch(
            self.url, 
            headers=self.header, 
            params=self.params, 
            data=self.data, 
            **kwargs
        )
        return data

    def cleanData(self):
        pass

  
class DeleteRequest(BaseRequest):

    def getData(self, **kwargs):
        data = self.client.delete(
            self.url, 
            headers=self.header, 
            params=self.params, 
            data=self.data, 
            **kwargs
        )
        return data

    def cleanData(self):
        pass


from setuptools import setup, find_packages

setup(
    name="sky-api-python-client",
    version='0.2.3',
    author='Warren Hyson',
    author_email='warren.hyson5@gmail.com',
    description="Package for communicating with the Blackbaud Sky API",
    url = "https://github.com/LearnThinkCreate/sky-api-python-client",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    test_suite='nose.collector',
)